# encoding: utf-8
from rapyuta_io.clients.api_client import CatalogConfig
from rapyuta_io.clients.package import Package
from rapyuta_io.clients.routed_network import RoutedNetwork
from rapyuta_io.clients.persistent_volumes import PersistentVolumes
from rapyuta_io.utils import RestClient, PackageNotFound, to_objdict, APIError, ConflictError
from rapyuta_io.utils.rest_client import HttpMethod
from rapyuta_io.utils.settings import CATALOG_API_PATH, VOLUME_PACKAGE_ID
from rapyuta_io.utils.utils import response_validator


class CatalogClient(CatalogConfig):

    def __init__(self, auth_token, project, catalog_api_host):
        super(CatalogClient, self).__init__(catalog_api_host, auth_token, project)
        self._api_path = CATALOG_API_PATH

    def _get_api_path(self):
        return self._catalog_api_host + self._api_path

    def _execute(self, url, method=HttpMethod.GET, retry_count=0, payload=None, query_params=None):
        rest_client = RestClient(url).method(method).headers(self._headers).retry(retry_count).query_param(query_params)
        response = rest_client.execute(payload)
        return response

    @response_validator(True, {404: PackageNotFound})
    def _get_service(self, package_id, retry_limit=0):
        url = self._get_api_path() + "?package_uid=%s" % package_id
        return self._execute(url, HttpMethod.GET, retry_limit)

    @response_validator(True)
    def get_all_packages(self, retry_limit):
        url = self._catalog_api_host + '/v2/catalog'
        return self._execute(url, HttpMethod.GET, retry_limit)

    def get_package(self, package_id, retry_limit):
        package = self._get_service(package_id, retry_limit)
        try:
            package = package['packageInfo']
        except Exception:
            raise APIError("packageInfo not present in the package")
        if package_id == VOLUME_PACKAGE_ID:
            pkg = PersistentVolumes(to_objdict(package))
            pkg.planId = 'default'
        else:
            pkg = Package(to_objdict(package), is_refreshed=True)
        setattr(pkg, 'packageId', package_id)
        setattr(pkg, 'is_refreshed', True)
        return pkg

    @response_validator(True)
    def get_deployment(self, deployment_id, retry_limit):
        path = '/serviceinstance/{}'.format(deployment_id)
        url = self._catalog_api_host + path
        return self._execute(url, HttpMethod.GET, retry_limit)

    @response_validator(True)
    def deployment_list(self, phases, retry_limit):
        url = self._catalog_api_host + '/deployment/list'
        query_params = None
        if phases:
            query_params = {'phase': map(str, phases)}
        return self._execute(url, HttpMethod.GET, retry_limit, query_params=query_params)

    @response_validator(True)
    def create_package(self, package_payload, retry_limit):
        url = self._catalog_api_host + '/serviceclass/add'
        return self._execute(url, HttpMethod.POST, retry_limit, package_payload)

    @response_validator(True)
    def create_routed_network(self, **network_payload):
        url = self._catalog_api_host + '/routednetwork'
        routed_network = self._execute(url, HttpMethod.POST, retry_count=0, payload=network_payload)
        return routed_network

    @response_validator(True)
    def get_routed_network(self, network_guid):
        url = self._catalog_api_host + '/routednetwork/{}'.format(network_guid)
        return self._execute(url, HttpMethod.GET)

    @response_validator(True)
    def delete_routed_network(self, network_guid):
        url = self._catalog_api_host + '/routednetwork/{}'.format(network_guid)
        return self._execute(url, HttpMethod.DELETE)

    @response_validator(True)
    def list_routed_network(self):
        url = self._catalog_api_host + '/routednetwork'
        return self._execute(url, HttpMethod.GET)
