# encoding: utf-8
import json
import os

from clients import DeviceManagerClient, _ParamserverClient
from rapyuta_io.clients.catalog_client import CatalogClient
from rapyuta_io.clients.core_api_client import CoreAPIClient
from rapyuta_io.clients.deployment import Deployment
from rapyuta_io.clients.device import Device
from rapyuta_io.clients.package import Package
from rapyuta_io.clients.persistent_volumes import VolumeInstance
from rapyuta_io.clients.routed_network import RoutedNetwork
from rapyuta_io.utils import to_objdict
from rapyuta_io.utils.error import ParameterMissingException
from rapyuta_io.utils.settings import VOLUME_PACKAGE_ID, default_host_config
from utils import InvalidAuthTokenException, InvalidParameterException
from rapyuta_io.clients.package import Runtime, ROSDistro, RestartPolicy


class Client(object):
    """
    Client class provides access to device, package, volume and deployment classes.

    """

    def __init__(self, auth_token, project):
        """
        Get new client object

        :param auth_token: Authentication token
        :type auth_token: string

        :param project: project_guid of the user
        :type project: string
        """

        super(Client, self).__init__()
        self._validate_auth_token(auth_token)
        self._catalog_client = CatalogClient(auth_token, project,
                                             catalog_api_host=self._get_api_endpoints('catalog_host'))
        self._core_api_client = CoreAPIClient(auth_token, project,
                                              core_api_host=self._get_api_endpoints('core_api_host'))
        self._dmClient = DeviceManagerClient(auth_token, project,
                                             device_api_host=self._get_api_endpoints('core_api_host'))
        self._paramserver_client = _ParamserverClient(auth_token, project, self._get_api_endpoints('core_api_host'))

    @staticmethod
    def _validate_auth_token(auth_token):
        if not auth_token:
            raise InvalidAuthTokenException("Authentication token is missing")

    @staticmethod
    def _get_api_endpoints(host_type):
        try:
            config_file = os.environ["RIO_CONFIG"]
            with open(config_file, 'r') as f:
                configuration = json.load(f)
            return configuration[host_type]
        except Exception:
            return default_host_config[host_type]

    def _add_auth_token(self, obj):
        setattr(obj, '_host', self._catalog_client._catalog_api_host)
        setattr(obj, '_auth_token', self._catalog_client._auth_token)
        setattr(obj, '_project', self._catalog_client._project)

    def get_persistent_volume(self, retry_limit=0):
        """
        Get persistent volume class

        :param retry_limit: Number of retry attempts to be carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: :py:class:`PersistentVolumes` class.
        :raises: :py:class:`APIError`: If the API returns an error, a status code
            of anything other than 200/201 is returned.

        Following example demonstrates how to get a persistent volume

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> persistent_volume = client.get_persistent_volume()

        """
        volume = self._catalog_client.get_package(VOLUME_PACKAGE_ID, retry_limit)
        self._add_auth_token(volume)
        return volume

    def get_package(self, package_id, retry_limit=0):

        """
        Get package class

        :param package_id: Package ID
        :type package_id: string
        :param retry_limit: Number of retry attempts to be carried out if any failures occurs \
                during the API call.
        :type retry_limit: int
        :return: an instance of :py:class:`~clients.catalog.Package` class.
        :raises: :py:class:`PackageNotFound`: If the given package id is not found.
        :raises: :py:class:`ComponentNotFoundException`: If the plan inside the package does
                  not have components.
        :raises: :py:class:`APIError`: If the API returns an error, a status code
            of anything other than 200/201 is returned.


        Following example demonstrates how to get the packages

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> package = client.get_package('test_package_id')

        """

        pkg = self._catalog_client.get_package(package_id, retry_limit)
        self._add_auth_token(pkg)
        return pkg

    def get_all_packages(self, retry_limit=0):

        """
        Get list of all packages created by the user.

        :param retry_limit: Number of retry attempts to be carried out if any failures occurs \
                during the API call.
        :type retry_limit: int
        :return: List of all packages. Each package is an instance of
                 :py:class:`~clients.catalog.Package` class.
        :raises: :py:class:`APIError`: If the API returns an error, a status code
            of anything other than 200/201 is returned.


        Following example demonstrates how to get the packages

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> package_list = client.get_all_packages()

        """

        package_list = self._catalog_client.get_all_packages(retry_limit)
        packages = list()
        for package in package_list['services']:
            if package['id'] == VOLUME_PACKAGE_ID:
                continue
            package = Package(to_objdict(package))
            self._add_auth_token(package)
            packages.append(package)
        return packages

    def get_deployment(self, deployment_id, retry_limit=0):
        """
        Get a deployment

        :param deployment_id: Deployment ID
        :type deployment_id: string
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: instance of class :py:class:`Deployment`:
        :raises: :py:class:`APIError`:  If the API returns an error, a status code of
            anything other than 200/201 is returned.



        Following example demonstrates how to get all the deployments.

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> deployment = client.get_deployment('deployment_id')

        """
        deployment = self._catalog_client.get_deployment(deployment_id, retry_limit)
        deployment = Deployment(to_objdict(deployment))
        self._add_auth_token(deployment)
        return deployment

    def get_volume_instance(self, instance_id, retry_limit=0):
        """
        Get a volume instance

        :param instance_id: Volume instance Id
        :type instance_id: string
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: instance of class :py:class:`VolumeInstance`:
        :raises: :py:class:`APIError`: If the API returns an error, a status code
            of anything other than 200/201 is returned.

        Following example demonstrates how to get a volume instance.

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> deployments = client.get_volume_instance('instance_id')

        """
        volume_instance = self._catalog_client.get_deployment(instance_id, retry_limit)
        volume_instance = VolumeInstance(to_objdict(volume_instance))
        self._add_auth_token(volume_instance)
        return volume_instance

    def get_all_deployments(self, phases=None, retry_limit=0):
        """
        Get all deployments created by the user.

        :param phases: optional parameter to filter out the deployments based on current deployment
        :type phases: list(DeploymentPhaseConstants)
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: list of instances of class :py:class:`Deployment`:
        :raises: :py:class:`APIError`: If the API returns an error, a status code
            of anything other than 200/201 is returned.

        Following example demonstrates how to get all the deployments.

            >>> from rapyuta_io import Client, DeploymentPhaseConstants
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> deployments = client.get_all_deployments()
            >>> deployments_list_filtered_by_phase = client.get_all_deployments(phases=
            >>>    [DeploymentPhaseConstants.SUCCEEDED, DeploymentPhaseConstants.PROVISIONING])

        """

        deployment_list = self._catalog_client.deployment_list(phases, retry_limit)
        deployments = list()
        for deployment in deployment_list:
            if deployment['packageId'] == VOLUME_PACKAGE_ID:
                continue
            deployment_object = Deployment(to_objdict(deployment))
            # todo: remove volume instances
            self._add_auth_token(deployment_object)
            deployments.append(deployment_object)
        return deployments

    def get_all_devices(self, online_device=False, arch_list=None, retry_limit=0):
        """
        Get all the devices

        :param online_device: The value True returns only those devices that are online,
            while the value False returns all devices
        :type online_device: bool
        :param arch_list: If set, another call is made to filter devices by architecture.
            Valid architectures can be found in :py:class:`~.DeviceArch` class. Note: Only online devices
            can be filtered by architecture. Therefore if `arch_list` is set, non-online
            devices are not included in the result irrespective of the value of `online_device`.
        :type arch_list: list
        :param retry_limit: No of retry attempts to be carried out if any failures occurs\
                during the API call.
        :type retry_limit: int
        :return: List of instances of :py:class:`~Device` class
        :raises: :py:class:`APIError`: If the API returns an error, a status code
            of anything other than 200/201 is returned

        Following example demonstrates how to get the device list

            >>> from rapyuta_io import Client
            >>> from rapyuta_io.clients import DeviceArch
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> devices = client.get_all_devices()
            >>> filtered_by_arch_devices = client.get_all_devices(arch_list=[
            >>>     DeviceArch.ARM32V7, DeviceArch.ARM64V8, DeviceArch.AMD64])

        """
        return self._dmClient.device_list(online_device, arch_list, retry_limit)

    def get_device(self, device_id, retry_limit=0):
        """
        Get information of a device.

        :param device_id: Device Id
        :type device_id: string
        :param retry_limit: No of retry attempts to be carried out if any failures occurs\
         during the API call.
        :type retry_limit: int
        :return: Instances of :py:class:`~Device` class.
        :raises: :py:class:`~utils.error.ResourceNotFoundError`: If the device with given device id is not present.
        :raises: :py:class:`APIError`: If the API returns an error, a status code of
            anything other than 200/201 is returned.

        Following example demonstrates how to select a device information.

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> device = client.get_device(device_id="device_id")

        """
        return self._dmClient.get_device(device_id, retry_limit)

    def create_package_from_manifest(self, manifest_filepath, retry_limit=0):
        """
        Create package from a manifest file

        :param manifest_filepath: File path of the manifest
        :type manifest_filepath: string
        :param retry_limit: No of retry attempts to be carried out if any failures occurs\
         during the API call.
        :type retry_limit: int
        :raises: :py:class:`~utils.error.ConflictError`: Package already exists.
        :raises: :py:class:`~utils.error.BadRequestError`: Invalid package details.
        :return: dict containing package details

        **Copy the below json to listener.json file** \n
        Following example demonstrates how to use create_package_from_manifest.

            .. code-block:: JSON

               { "packageVersion": "v1.0.0", "plans": [{"includePackages": [], "singleton": false, "name": "default",
               "inboundROSInterfaces": {"services": [], "topics": [], "actions": []}, "dependentDeployments": [],
               "components": [{"executables": [{"cmd": ["roslaunch listener listener.launch"], "name": "listenerExec"}],
               "name": "default", "parameters": [], "architecture": "arm32v7", "requiredRuntime": "device",
               "ros": {"services": [], "topics": [], "isROS": true, "actions": []}, "description": ""}],
               "exposedParameters": [], "metadata": {}}], "name": "listener", "apiVersion": "v1.0.0",
               "description": "listener arm32v7 sdk test package" }

        >>> from rapyuta_io import Client
        >>> client = Client(auth_token='auth_token', project='project_guid')
        >>> package_details = client.create_package_from_manifest('listener.json')

        """
        manifest = self._get_manifest_from_file(manifest_filepath)
        return self._catalog_client.create_package(manifest, retry_limit)

    @staticmethod
    def _get_manifest_from_file(manifest_filepath):
        with open(manifest_filepath, 'r') as f:
            return json.load(f)

    def create_package(self, manifest, retry_limit=0):
        """
        Create package from manifest dict.

        :param manifest: dict containing package details
        :type manifest: dict
        :param retry_limit: No of retry attempts to be carried out if any failures occurs\
         during the API call.
        :type retry_limit: int
        :raises: :py:class:`~utils.error.ConflictError`: Package already exists.
        :raises: :py:class:`~utils.error.BadRequestError`: Invalid package details.
        :return: dict containing package details

        Following example demonstrates how to use create_package.

            >>> manifest = {'packageVersion': 'v1.0.0',
                 'plans': [{'includePackages': [],
                 'singleton': False,
                 'name': 'default',
                 'inboundROSInterfaces': {'services': [], 'topics': [], 'actions': []},
                 'dependentDeployments': [],
                 'components': [{'executables': [{'cmd': ['roslaunch listener listener.launch'],
                     'name': 'listenerExec'}],
                     'name': 'default',
                     'parameters': [],
                     'architecture': 'arm32v7',
                     'requiredRuntime': 'device',
                     'ros': {'services': [], 'topics': [], 'isROS': True, 'actions': []},
                     'description': ''}],
                     'exposedParameters': [],
                     'metadata': {}}],
                 'name': 'listener',
                 'apiVersion': 'v1.0.0',
                 'description': 'listener arm32v7 sdk test package'}
            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> package_details = client.create_package(manifest)

        """
        return self._catalog_client.create_package(manifest, retry_limit)

    def upload_configurations(self, rootdir, tree_names=None, delete_existing_trees=False):
        """
        Traverses rootdir and uploads configurations following the same directory structure.

        :param rootdir: Path to directory containing configurations
        :type rootdir: str
        :param tree_names: List of specific configuration trees to upload. If None, all trees under rootdir are uploaded
        :type tree_names: list[str], optional
        :param delete_existing_trees: For each tree to upload, delete existing tree at the server. Defaults to False
        :type delete_existing_trees: bool, optional

        Following example demonstrates how to use upload_configurations and handle errors.

            >>> from rapyuta_io import Client
            >>> from rapyuta_io.utils.error import BadRequestError, InternalServerError
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> try:
            ...     client.upload_configurations('path/to/configs/source_dir',
            ...                                  tree_names=['config_tree1', 'config_tree2'],
            ...                                  delete_existing_trees=True)
            ... except (BadRequestError, InternalServerError) as e:
            ...     print 'failed API request', e.tree_path, e
            ... except (IOError, OSError) as e:
            ...     print 'failed file/directory read', e

        """
        return self._paramserver_client.upload_configurations(rootdir, tree_names, delete_existing_trees)

    def download_configurations(self, rootdir, tree_names=None, delete_existing_trees=False):
        """
        Download all configurations to rootdir following the same directory structure. If rootdir does not exist, it is
        created.

        :param rootdir: Path to directory to store downloaded configurations
        :type rootdir: str
        :param tree_names: List of specific configuration trees to download. If None, all trees are downloaded
        :type tree_names: list[str], optional
        :param delete_existing_trees: For each tree to download, delete existing tree on the filesystem. Defaults to
            False
        :type delete_existing_trees: bool, optional

        Following example demonstrates how to use download_configurations and handle errors.

            >>> from rapyuta_io import Client
            >>> from rapyuta_io.utils.error import APIError, InternalServerError
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> try:
            ...     client.download_configurations('path/to/destination_dir',
            ...                                    tree_names=['config_tree1', 'config_tree2'],
            ...                                    delete_existing_trees=True)
            ... except (APIError, InternalServerError) as e:
            ...     print 'failed API request', e.tree_path, e
            ... except (IOError, OSError) as e:
            ...     print 'failed file/directory creation', e

        """
        return self._paramserver_client.download_configurations(rootdir, tree_names, delete_existing_trees)

    def apply_parameters(self, device_list, tree_names=None, retry_limit=0):
        """
        Applies configuration parameters for the given device_list. If tree_names is given, only these trees are
        applied.

        :param device_list: List of device IDs
        :type device_list: list[str]
        :param tree_names: List of configuration tree names
        :type tree_names: list[str]
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: List of dictionaries - each with device_id, bool success status, and an error message if the success
            status is False for that device_id
        :rtype: list[dict]

        Following example demonstrates how to use apply_parameters and handle errors.

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> devices = client.get_all_devices()
            >>> response = client.apply_parameters([device.deviceId for device in devices])
            >>> for device in response:
            ...     if not device['success']:
            ...         print device['device_id'], device['error']

        """
        return self._dmClient.apply_parameters(device_list, tree_names, retry_limit)

    def get_all_static_routes(self):
        """
        List all static routes in a project.

        :return: Instances of :py:class:`~StaticRoute` class.

        Following example demonstrates how to list all static routes

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> device = client.get_all_static_routes()
        """
        return self._core_api_client.get_all_static_routes()

    def get_static_route(self, route_guid):
        """
        Get static routes by its guid

        :param route_guid: GUID string of a :py:class:`~StaticRoute` class.
        :type route_guid: str
        :return: Instance of :py:class:`~StaticRoute` class.

        Following example demonstrates how to get a static route

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> static_route_guid = client.get_all_static_routes()[0]['guid']
            >>> device = client.get_static_route(static_route_guid)
        """
        return self._core_api_client.get_static_route(route_guid)

    def get_static_route_by_name(self, name):
        """
        Get static routes by its name

        :param name: Name (urlPrefix) of the :py:class:`~StaticRoute` instance.
        :type name: str
        :return: Instance of :py:class:`~StaticRoute` class or None if it doesn't exist

        Following example demonstrates how to get a static route by its name/url prefix

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> static_route = client.get_static_route_by_name('example-route')
        """
        return self._core_api_client.get_static_route_by_name(name)

    def create_static_route(self, name):
        """
        Create static route of a certain name

        :param name: Name of the static route. It should follow ^[a-z][a-z0-9-]*$ and should not contain black listed keyword and be of length between 4 and 64
        :type name: str
        :return: Instance of :py:class:`~StaticRoute` class.

        Following example demonstrates how to create a static route.

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> device = client.create_static_route('example-route')
        """
        return self._core_api_client.create_static_route(name)

    def get_routed_network(self, network_guid):
        """
        Get routed network for the guid

        :param network_guid: guid of routed network
        :type  network_guid: str
        :return: Instance of :py:class:`~rapyuta_io.clients.routed_network.RoutedNetwork` class.

            >>> from rapyuta_io import Client
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> routed_network = client.get_routed_network(network_guid)

        """
        routed_network = self._catalog_client.get_routed_network(network_guid)
        routed_network = RoutedNetwork(to_objdict(routed_network))
        routed_network.phase = routed_network.internalDeploymentStatus.phase
        routed_network.status = routed_network.internalDeploymentStatus.status
        routed_network.error_code = routed_network.get_error_code()
        self._add_auth_token(routed_network)
        return routed_network

    def get_all_routed_networks(self):
        """
        List routed network

        :return: List instance of :py:class:`~rapyuta_io.clients.routed_network.RoutedNetwork` class.
        """

        routed_networks = []
        networks = self._catalog_client.list_routed_network()
        for routed_network in networks:
            routed_network = RoutedNetwork(to_objdict(routed_network))
            internal_deployment_status = routed_network.internalDeploymentStatus
            routed_network.phase = internal_deployment_status.phase
            routed_network.error_code = routed_network.get_error_code()
            self._add_auth_token(routed_network)
            routed_networks.append(routed_network)
        return routed_networks

    def create_cloud_routed_network(self, name, ros_distro, shared):

        """
        Create a routed network

        :param name: Name of the routed network.
        :type name: str
        :param ros_distro: ros ditro of the runtime.
        :type ros_distro: enum :py:class:`~rapyuta_io.clients.package.ROSDistro`
        :param shared: Whether the network should be shared.
        :type shared: bool
        :return: Instance of :py:class:`~rapyuta_io.clients.routed_network.RoutedNetwork` class.

        Following example demonstrates how to create a routed network.

            >>> from rapyuta_io import Client
            >>> from rapyuta_io.clients.package import ROSDistro
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> routed_network = client.create_cloud_routed_network('network_name', ROSDistro.KINETIC, True)
        """

        if ros_distro not in ROSDistro:
            raise InvalidParameterException('ROSDistro must be one of rapyuta_io.clients.package.ROSDistro')

        routed_network = self._catalog_client.create_routed_network(name=name, runtime=Runtime.CLOUD,
                                                                    rosDistro=ros_distro,
                                                                    shared=shared, parameters={})
        routed_network = RoutedNetwork(to_objdict(routed_network))
        return self.get_routed_network(routed_network.guid)

    def create_device_routed_network(self, name, ros_distro, shared,
                                     device, network_interface, restart_policy=RestartPolicy.Always):

        """
        Create a routed network

        :param name: Name of the routed network.
        :type name: str
        :param ros_distro: ros ditro of the runtime.
        :type ros_distro: enum :py:class:`~rapyuta_io.clients.package.ROSDistro`
        :param shared: Whether the network should be shared.
        :type shared: bool
        :param device: device on which the routed network is deployed.
        :type device: Instance of :py:class:`~Device` class.
        :param network_interface: network interface to which routed network is binded.
        :type network_interface: str
        :param restart_policy: restart policy of routed network.
        :type restart_policy: enum :py:class:`~rapyuta_io.clients.package.RestartPolicy`
        :return: Instance of :py:class:`~rapyuta_io.clients.routed_network.RoutedNetwork` class.

        Following example demonstrates how to create a routed network.

            >>> from rapyuta_io import Client
            >>> from rapyuta_io.clients.package import ROSDistro
            >>> client = Client(auth_token='auth_token', project='project_guid')
            >>> routed_network = client.create_device_routed_network('network_name',
            >>>                                                      ROSDistro.KINETIC, True)
        """

        parameters = {}
        if ros_distro not in ROSDistro:
            raise InvalidParameterException('ROSDistro must be one of rapyuta_io.clients.package.ROSDistro')

        if not isinstance(device, Device):
            raise InvalidParameterException('device must be of type rapyuta_io.clients.device.Device')

        ip_interfaces = device.ip_interfaces or {}
        if network_interface not in ip_interfaces.keys():
            raise InvalidParameterException('NETWORK_INTERFACE should be in {}'.format(ip_interfaces.keys()))

        if restart_policy not in RestartPolicy:
            raise InvalidParameterException('RestartPolicy must be one of rapyuta_io.clients.package.RestartPolicy')

        parameters['device_id'] = device.uuid
        parameters['NETWORK_INTERFACE'] = network_interface
        parameters['restart_policy'] = restart_policy

        routed_network = self._catalog_client.create_routed_network(name=name, runtime=Runtime.DEVICE,
                                                                    rosDistro=ros_distro,
                                                                    shared=shared, parameters=parameters)
        routed_network = RoutedNetwork(to_objdict(routed_network))
        return self.get_routed_network(routed_network.guid)
